#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libgen.h>
#include "iriver.h"
#include <ftw.h>

#define IFP_REQ_TYPE              0xc0
#define IFP_GET_CAPACITY_REQUEST  0x14
#define IFP_GET_FREE_REQUEST      0x15
#define IFP_SEND_DATA             0x18
#define IFP_LS                    0x0f
#define IFP_LS_NEXT               0x10
#define IFP_BULK_TO               0x03
#define IFP_BULK_FROM             0x83
#define IFP_LS_LEAVE              0x11
#define IFP_GET_FILE_INFO_SIZE    0x0b
#define IFP_FILE_INFO             0x05
#define IFP_FILE_LEAVE            0x0d
#define IFP_FILE_UPLOAD           0x06
#define IFP_FILE_DOWNLOAD         0x07
#define IFP_FILE_UPLOAD_APPEND    0x08
#define IFP_FILE_DELETE           0x0e
#define IFP_DIR_DELETE            0x13
#define IFP_DIR_CREATE            0x12
#define IFP_02_COMMAND            0x02
#define IFP_02_BATTERY            0x08
#define IFP_02_STRING             0x00
#define IFP_02_FIRMWARE           0x03
#define IFP_FORMAT                0x16
#define IFP_FIRMWARE_UPDATE       0x17

#define TOUT  10000
#define R_TIME 10000
#define TRUE  1
#define FALSE 0

#define usb_req(req_type,i,j,name)  { \
	usb_control_msg(dh,IFP_REQ_TYPE,req_type,i,j,name,sizeof(name),TOUT); \
	if (!ifp_ok(name)) return IRIVER_GENERIC_USB_ERROR; }

/*#define DEBUG_LIBIRIVER
#define DEBUG_LIBIRIVER_VERBOSE*/

#ifdef DEBUG_LIBIRIVER
#define DEBUG printf
#else 
#define DEBUG
#endif

#ifdef DEBUG_LIBIRIVER_VERBOSE
#define DEBUG2 printf
#else
#define DEBUG2
#endif


static int string2unicode(char *dest, const char *src)
{
	int i = 0, j = 0;
  
	while (src[j] != 0) {
		dest[i++] = src[j++];
		dest[i++] = 0;
	}
	dest[i++] = 0;
	dest[i++] = 0;
	return i;
}
                                                                                                                
static int unicode2string(char *dest, const char *src) 
{
	int i = 0, j = 0;
  
	while ((src[j] != 0) || (src[j+1] != 0)) {
		dest[i++] = src[j];
		j += 2;
	}
	dest[i++] = 0;
	return i;
}

static int slash2backslash(char *dest,const char *src) 
{
	int i = 0;
  
	while (src[i] != 0) 
		if (src[i] == '/') 
			dest[i++] = '\\'; 
		else 
			dest[i] = src[i++];
	dest[i] = 0;
	return i;
}

static int ifp_ctl_ret(unsigned char *control) 
{
	return (control[0] + 256*control[1]);
}

static int ifp_ok(unsigned char *control) 
{
	return (ifp_ctl_ret(control) == 1);
}

int ifp_send_data(usb_dev_handle *dh, char *data, size_t size) 
{
	unsigned char ctl[8];

	usb_req(IFP_SEND_DATA,size,0,ctl);
	return (usb_bulk_write(dh,IFP_BULK_TO,data,size,TOUT));
}

static int ifp_send_filename(usb_dev_handle *dh, const char *name) 
{
	unsigned char buf[0x400];
	unsigned char strbuf[0x200];
  
	slash2backslash(strbuf,name);
	string2unicode(buf,strbuf);
	return ifp_send_data(dh,buf,sizeof(buf));
}

static int file_has_extension(const char *str, const char *suffix)
{
	return (strcasecmp (str + strlen (str) - strlen (suffix), suffix) == 0);
}

static int file_has_basename(const char *str, const char *base)
{
	return ((strstr(str,base) == str+strlen(str)-strlen(base)));
}


/**
 * ifp_get_capacity:
 * @dh: a usb_dev_handle
 * 
 * Returns: The total space of the iRiver expressed in bytes
 **/
int ifp_get_capacity(usb_dev_handle *dh) 
{
	unsigned char ctl[8];

	DEBUG2("ifp_get_capacity(): begin\n");
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_CAPACITY_REQUEST,0,0, 
			(char *)&ctl, sizeof(ctl),TOUT); 
	if (!ifp_ok(ctl)) 
		return IRIVER_GENERIC_USB_ERROR;    
	DEBUG2("ifp_get_capacity(): end\n");
	return (ctl[7] << 24) + (ctl[6] << 16) + (ctl[5] << 8) + ctl[4];
}

/**
 * ifp_get_free:
 * @dh: a usb_dev_handle
 * 
 * Returns: The available space of the iRiver expressed in bytes
 **/
int ifp_get_free(usb_dev_handle *dh) 
{
	unsigned char ctl[8];
	DEBUG2("ifp_get_free(): begin\n");
	usb_req(IFP_GET_FREE_REQUEST,0,0,ctl);
	DEBUG2("ifp_get_free(): end\n");
	return (ctl[7] << 24) + (ctl[6] << 16) + (ctl[5] << 8) + ctl[4];
}

/**
 * ifp_file_exists:
 * @dh: a usb_dev_handle
 * @name: name of the file to be tested for existance
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_FILE_NOT_FOUND if @name can't be found,
 * %IRIVER_CANNOT_SEND_FILENAME if @name can't be even sent to the iRiver
 **/
int ifp_file_exists(usb_dev_handle *dh, const char *name)
{
	unsigned char ctl[4];
	unsigned char ctl8[8];

	DEBUG2("ifp_file_exists(): begin -- name:%s\n",name);
	if (ifp_send_filename(dh,name) < 0 ) 
		return IRIVER_CANNOT_SEND_FILENAME;
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS,0,0,ctl,sizeof(ctl),TOUT);
	if (ifp_ok(ctl) == 1) {
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
		DEBUG2("ifp_file_exists(): %s is a directory. end\n", name);
		return IRIVER_DIRECTORY_EXISTS;
	}
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_INFO,1,0,ctl,sizeof(ctl),TOUT);
	if (!ifp_ok(ctl)) {
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
		DEBUG2("ifp_file_exists(): %s cannot be found. end\n", name);
		return IRIVER_FILE_NOT_FOUND;
	} 
	usb_req(IFP_GET_FILE_INFO_SIZE,0,0,ctl8);
	usb_req(IFP_FILE_LEAVE,0,0,ctl);
	DEBUG2("ifp_file_exists(): %s is a file. end.\n", name);
	return IRIVER_FILE_EXISTS;
}

static void ifp_get_file_info(usb_dev_handle * dh, const char *name, ifp_file_info **list)
{
	unsigned char ctl[4];
	unsigned char ctl8[8];
	int size;    

	ifp_send_filename(dh,name);
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_INFO,1,0,ctl,sizeof(ctl),TOUT);
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_FILE_INFO_SIZE,0,0,ctl8,sizeof(ctl8),TOUT);
	size = ctl8[4] + (ctl8[5] << 8) + (ctl8[6] << 16) + (ctl8[7] << 24);
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
	/* FIXME: this one needs to be sorted out a little better...*/
	(*list)->name = strdup(basename(strdup(name)));
	(*list)->is_file = TRUE;
	(*list)->size = size;
	(*list)->next = NULL;
	/* Somewhat broken mime-type detection goes here... */
	if (file_has_extension(name,"mp3")) 
		(*list)->mimetype = strdup(MP3_MIMETYPE);
	else if (file_has_extension(name,"ogg")) 
		(*list)->mimetype = strdup(OGG_MIMETYPE);
	else if (file_has_extension(name,"wma")) 
		(*list)->mimetype = strdup(WMA_MIMETYPE);
	else if (file_has_extension(name,"asf")) 
		(*list)->mimetype = strdup(ASF_MIMETYPE);
	else 
		(*list)->mimetype = strdup(DEFAULT_MIMETYPE);
}

static void ifp_get_directory_info(usb_dev_handle *dh, const char *name, ifp_file_info **list)
{
	(*list)->name = strdup(name);
	(*list)->mimetype = strdup(DIRECTORY_MIMETYPE);
	(*list)->size = 0;
	(*list)->is_file = FALSE;
	(*list)->next = NULL;
}

/**
 * ifp_get_info:
 * @dh: a usb_dev_handle
 * @name: name of the file or directory to retrieve the info about
 * @p: #ifp_file_info structure that will hold the information
 * 
 * Function to retrieve information about a file. (size, mime, filetype)
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_FILE_NOT_FOUND if @name can't be found,
 * %IRIVER_CANNOT_SEND_FILENAME if @name can't be even sent to the iRiver
 **/
int ifp_get_info(usb_dev_handle *dh, const char *name, ifp_file_info **p)
{
	unsigned char ctl[4];
	ifp_file_info *list = malloc(sizeof(ifp_file_info));

	DEBUG2("ifp_get_info(): begin -- name=%s\n",name);
	*p = list;
	list->name = NULL;  
	if (ifp_file_exists(dh,name) == IRIVER_FILE_NOT_FOUND) 
		return IRIVER_FILE_NOT_FOUND;
	if (ifp_send_filename(dh,name) < 0 ) 
		return IRIVER_CANNOT_SEND_FILENAME;
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS,0,0,ctl,sizeof(ctl),TOUT);
	if (!ifp_ok(ctl)) 
		ifp_get_file_info(dh, name, &list);
	else
		ifp_get_directory_info(dh, name, &list);
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
	list->next = NULL;
	DEBUG2("ifp_get_info(): end\n");
	return 0;
}


static void ifp_alloc_file_info(ifp_file_info **fi)
{
	ifp_file_info *fi_next;
  
	fi_next = (ifp_file_info *) malloc(sizeof(ifp_file_info));
	(*fi)->next = fi_next;
	*fi = fi_next;
}

/**
 * ifp_ls:
 * @dh: a usb_dev_handle
 * @name: name of the directory to retrieve a listing for
 * @p: the list that will contain all of the directory entries
 * 
 * This function retrieves all files in the directory specified by @name and
 * fills in the @p parameter with the infos about every entry 
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_CANNOT_SEND_FILENAME if problems with 
 * the iRiver occurred, %IRIVER_DIRECTORY_NOT_FOUND if @name is not a
 * directory or does not exist.
 *
 **/
int ifp_ls (usb_dev_handle * dh, const char *name, ifp_file_info ** p)
{
	int ret=0;
	ifp_file_info *list = NULL;
	unsigned char ctl[4];
	unsigned char ctl8[8];
	unsigned char inbuf[0x100];
	unsigned char strbuf[0x200];
	unsigned char strbuf2[0x200];
	int size;
		
	DEBUG("ifp_ls(): begin -- name=%s\n",name);
	if (ifp_file_exists(dh,name) != IRIVER_DIRECTORY_EXISTS) 
		return IRIVER_DIRECTORY_NOT_FOUND;
	/* Should *p be zeroed out before this..? Yes ? */
	if (ifp_empty_dir(dh, name) == TRUE) {
		*p = NULL;
		DEBUG2("ifp_ls(): directory is empty. end.\n");
		return IRIVER_OK;
	}
	list = malloc(sizeof(ifp_file_info));
	list->next = NULL;
	*p = list;
	list->size = -1;
	if (ifp_send_filename(dh,name) < 0) 
		return IRIVER_CANNOT_SEND_FILENAME;
	usb_req(IFP_LS,0,0,ctl);
	do {
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_NEXT,3,0,ctl,sizeof(ctl),TOUT);  // 3=magic !!!??? 
		ret = ifp_ctl_ret(ctl);
		switch (ret) {
		case 0:
			goto LSEND;
			break;
		case 1: /* File */
			usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
			unicode2string(strbuf,inbuf);
			strcpy(strbuf2,name);
			if (strcmp(name,"/") != 0) 
				strcat(strbuf2,"/");
			strcat(strbuf2,strbuf); // path/file
			if (list->size != -1) 
				ifp_alloc_file_info(&list);
			ifp_get_file_info(dh, strbuf2, &list);
			break;
		case 2: /* Directory */
			usb_bulk_read(dh,IFP_BULK_FROM,inbuf,sizeof(inbuf),TOUT);
			unicode2string(strbuf,inbuf);
			if (list->size != -1) 
				ifp_alloc_file_info(&list);
			ifp_get_directory_info(dh, strbuf, &list);
			break;
		default:
			goto LSEND;
			break;
		}
	} while (ret!=0);
LSEND:
//  list->next=NULL;
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
	DEBUG("ifp_ls(): end\n");
	return IRIVER_OK;
}

/**
 * ifp_upload_file:
 * @dh: a usb_dev_handle
 * @iname: name of the file on the iRiver
 * @name: name of the local file to upload
 * @progress: callback function to signal progress. Set to NULL if no progress
 * indication is needed
 * 
 * Uploads @name on the iRiver where it will be @iname
 * 
 * Returns: %IRIVER_OK if foo successful, %IRIVER_NO_FREE_SPACE if there 
 * is not enough free space to upload the file, %IRIVER_LOCAL_FILE_OPEN error 
 * if problems opening the local file occurred, %IRIVER_CANNOT_SEND_FILENAME 
 * if there were problems communicating with the iRiver,
 * %IRIVER_DIRECTORY_EXISTS if @name exists and is a directory, 
 * %IRIVER_FILE_EXISTS if @name exists and is a file.
 **/
int ifp_upload_file(usb_dev_handle *dh, const char *iname, const char *name, ifp_progress progress) 
{
	unsigned char buf[0x4000];
	unsigned char ctl[4]; 
	int i, filesize = 0;
	struct stat st;
	FILE *f;

	DEBUG("ifp_upload_file(): begin -- iname=%s name=%s\n",iname, name);
	if ((i = ifp_file_exists(dh,iname)) != IRIVER_FILE_NOT_FOUND) 
		return i;
	if ((f = fopen(name,"rb")) == NULL) 
		return IRIVER_LOCAL_FILE_OPEN_ERROR;
	stat(name,&st);
	if ((filesize = st.st_size) > ifp_get_free(dh)) 
		return IRIVER_NO_FREE_SPACE;
	if (ifp_send_filename(dh,iname) < 0 ) 
		return IRIVER_CANNOT_SEND_FILENAME;
	usb_req(IFP_FILE_UPLOAD,filesize,0,ctl);
	while (!feof(f)) {
		if (progress != NULL)
			progress(iname);
		i = fread(buf,1,sizeof(buf),f);
		ifp_send_data(dh,buf,i);
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_UPLOAD_APPEND,i,0,ctl,sizeof(ctl),TOUT);
	}
	ifp_send_data(dh,buf,0);
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_UPLOAD_APPEND,0,0,ctl,sizeof(ctl),TOUT);
	usb_req(IFP_FILE_LEAVE,0,0,ctl);
	fclose(f);
	DEBUG("ifp_upload_file(): end\n");
	return IRIVER_OK;
}

/**
 * ifp_delete_file:
 * @dh: a usb_dev_handle
 * @name: name of the file to be removed
 * @progress: callback function to signal progress. Set to NULL if no progress
 * indication is needed
 * 
 * Removes a file from the iRiver
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_FILE_NOT_FOUND if the file
 * doesn't exist, %IRIVER_CANNOT_SEND_FILE_NAME if there were problems 
 * communicating with the iRiver.
 **/
int ifp_delete_file(usb_dev_handle *dh, const char *name, ifp_progress progress)
{
	unsigned char ctl[4];
	int i;
  	
	DEBUG("ifp_delete_file(): begin -- name=%s\n",name);
	if (ifp_file_exists(dh,name)!=IRIVER_FILE_EXISTS) 
		return IRIVER_FILE_NOT_FOUND;
	if (ifp_send_filename(dh,name) < 0 ) 
		return IRIVER_CANNOT_SEND_FILENAME;
	usb_req(IFP_FILE_DELETE,0,0,ctl);
	printf("inside delete_file %x\n", progress);
	if (progress != NULL)
		progress(name);
	DEBUG("ifp_delete_file(): end\n");
	return IRIVER_OK;
}

/**
 * ifp_empty_dir:
 * @dh: a usb_dev_handle
 * @name: name of a directory
 * 
 * Checks to see if a directory is empty or not
 * 
 * Returns: %TRUE if empty, %FALSE if not or %IRIVER_DIRECTORY_NOT_FOUND if
 * the dir is not found and %IRIVER_CANNOT_SEND_FILENAM if there were problems
 * communicating with the iRiver
 *
 **/
int ifp_empty_dir(usb_dev_handle *dh, const char *name) 
{
	int i;
	unsigned char ctl[4];
  
	if (ifp_file_exists(dh,name) != IRIVER_DIRECTORY_EXISTS) 
		return IRIVER_DIRECTORY_NOT_FOUND;
	if (ifp_send_filename(dh,name) < 0 ) 
		return IRIVER_CANNOT_SEND_FILENAME;
	usb_req(IFP_LS,0,0,ctl);
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_NEXT,3,0,ctl,sizeof(ctl),TOUT);  // 3=magic !!!???
	i = (ifp_ctl_ret(ctl) == 0) ? TRUE : FALSE;
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_LS_LEAVE,0,0,ctl,sizeof(ctl),TOUT);
	return i;
}

/**
 * ifp_model_name:
 * @dh: a usb_dev_handle
 * @string: name of the string
 * 
 * Fills @string with the model name of the iRiver
 * 
 * Returns: %IRIVER_OK 
 **/
int ifp_model_name(usb_dev_handle *dh, char *string) 
{
	unsigned char ctl8[8];
	unsigned char bbuf[32];

	DEBUG2("ifp_model_name(): begin\n");
	usb_req(IFP_02_COMMAND,0,IFP_02_STRING,ctl8);
	usb_bulk_read(dh,IFP_BULK_FROM,bbuf,sizeof(bbuf),TOUT);
	strncpy(string,bbuf,32);
	DEBUG2("ifp_model_name(): end\n");
	return IRIVER_OK;
}

static char * ifp_compose_name(const char *str1, const char *str2)
{
	char *tmp;

	tmp=(char *) malloc(strlen(str1)+strlen(str2)+2);
	strcpy(tmp, str1);
	strcat(tmp, "/");
	strcat(tmp, str2);
	return tmp;
}



/**
 * ifp_delete_dir:
 * @dh: a usb_dev_handle
 * @name: name of the directory to be removed
 * @progress: callback function to signal progress. Set to NULL if no progress
 * indication is needed
 * 
 * Removes a directory. The directory must be empty
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_DIRECTORY_NOT_FOUND if @name is
 * not found or is a regular file, %IRIVER_DIRECTORY_NOT_EMPTY if the
 * directory is not empty, %IRIVER_CANNOT_SEND_FILENAME in case of other I/O
 * problems. 
 **/
int ifp_delete_dir(usb_dev_handle *dh, const char *name, ifp_progress progress) 
{
	unsigned char ctl[4];
	int r;
        ifp_file_info *list=NULL, *f=NULL, *l;
	char *iname;
	
	DEBUG("ifp_delete_dir(): begin -- name%s\n",name);
	if (ifp_file_exists(dh,name) != IRIVER_DIRECTORY_EXISTS) 
		return IRIVER_DIRECTORY_NOT_FOUND;
	if (ifp_empty_dir(dh, name)) 
	{
		DEBUG2("ifp_delete_dir(): directory is empty.\n");
		if (ifp_send_filename(dh,name) < 0 ) 
			return IRIVER_CANNOT_SEND_FILENAME;
		usb_req(IFP_DIR_DELETE,0,0,ctl);
		DEBUG2("ifp_delete_dir(): empty directory deleted. end.\n");
		return IRIVER_OK;
	}
	DEBUG2("ifp_delete_dir(): ifp_ls()\n");
	if ((r = ifp_ls(dh, name, &list))!=IRIVER_OK) return r;
	f = list;
	while (list != NULL) 
	{ 
		iname = ifp_compose_name(name, list->name);
		if (list->is_file)
		{	
			DEBUG2("ifp_delete_dir(): ifp_delete_file(): iname=%s\n",iname);
			r = ifp_delete_file(dh,iname, progress);
			if (r == IRIVER_FILE_NOT_FOUND)
			{
				DEBUG2("\033[1mifp_delete_dir(): ifp_delete_file() returns r=%d.\033[0m\n", r);
				DEBUG2("ifp_delete_dir(): bogus call to ifp_get_info()\n");
				ifp_get_info(dh, iname, &l);
				DEBUG2("ifp_delete_dir(): retry ifp_delete_file()\n");
				r = ifp_delete_file(dh,iname, progress);
				DEBUG2("ifp_delete_dir(): ifp_delete_file() returns r=%d\n",r);
			}
			if (r != IRIVER_OK) return r;

		}
		else
		{
			DEBUG2("ifp_delete_dir(): ifp_delete_dir() -- iname=%s\n",iname);
			r = ifp_delete_dir(dh,iname, progress);
			if ((r == IRIVER_DIRECTORY_NOT_FOUND) || (r==IRIVER_FILE_NOT_FOUND))
			{
				DEBUG2("\033ifp_delete_dir(): ifp_delete_dir() returns r=%d. retrying.\033[0m\n", r);
				sleep(2);
				r = ifp_delete_dir(dh,iname, progress);
				DEBUG2("ifp_delete_dir(): ifp_delete_dir() returns r=%d\n");
			}
			if (r != IRIVER_OK) return r;
		}
		free(iname);
		list = list->next;
	}
	ifp_free_file_list(&f);
	if ((r = ifp_delete_dir(dh, name, progress)) != IRIVER_OK) return r;
	DEBUG("ifp_delete_dir(): end\n");
	return IRIVER_OK;	
}

/**
 * ifp_make_dir:
 * @dh: a usb_dev_handle
 * @name: name of the directory to be created
 * 
 * Creates a directory on the iRiver
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_FILE_EXISTS or
 * %IRIVER_DIRECTORY_EXISTS if @name exists and %IRIVER_CANNOT_SEND_FILENAME
 * if I/O problems occurred.
 **/
int ifp_make_dir(usb_dev_handle *dh, const char *name) 
{
	unsigned char ctl[4];
	int i;
	
	DEBUG("ifp_make_dir(): begin -- name=%s\n",name);
	if (i = ifp_file_exists(dh,name) != IRIVER_FILE_NOT_FOUND) 
		return i;
	if (ifp_send_filename(dh,name) < 0 ) 
		return IRIVER_CANNOT_SEND_FILENAME;
	usb_req(IFP_DIR_CREATE,0,0, ctl);
	DEBUG("ifp_make_dir(): end\n");
	return IRIVER_OK;
}

/**
 * ifp_format:
 * @dh: a usb_dev_handle
 * 
 * Formats the iRiver completely
 * 
 * Returns: %IRIVER_OK
 *
 **/
int ifp_format(usb_dev_handle *dh) 
{
	unsigned char ctl[4];
  	
	DEBUG("ifp_format(): begin\n");
	usb_req(IFP_FORMAT, 0, 0, ctl);
	DEBUG("ifp_format(): end\n");
	return IRIVER_OK;
}

/**
 * ifp_firmware_update:
 * @dh: a usb_dev_handle
 * @name: name of the firmware file to upgrade the iRiver to
 * 
 * Upgrades the iRiver's firmware. 
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_UPLOAD_ERROR if 
 * the upload of the firmware did not work, %IRIVER_FILE_EXISTS if the
 * firmware file is already on the iRiver, %IRIVER_FIRMWARE_BASENAME_INCORRECT
 * if the firmware file is not correct.
 **/
int ifp_firmware_update(usb_dev_handle **dh, char *name) 
{
	int i;
	unsigned char ctl[4];
	char *bn;
	
	DEBUG("ifp_firmware_update(): begin -- name=%s\n");
	if (file_has_basename(name, FIRMWARE_BASENAMETC)) 
		bn = strdup(FIRMWARE_BASENAMETC);
	else if (file_has_basename(name, FIRMWARE_BASENAMET)) 
		bn = strdup(FIRMWARE_BASENAMET);
        else
          return IRIVER_FIRMWARE_BASENAME_INCORRECT; 
	if ((i = ifp_file_exists(*dh,bn)) != IRIVER_FILE_NOT_FOUND) 
		return i;
	DEBUG2("ifp_firmware_update(): ifp_file_upload()\n");
	if (ifp_upload_file(*dh, bn, name, NULL) != IRIVER_OK) /* No progress on firmware update anyway */
		return IRIVER_UPLOAD_ERROR;
	usb_control_msg(*dh, IFP_REQ_TYPE, IFP_FIRMWARE_UPDATE, 0, 0, ctl, sizeof(ctl), TOUT);
	DEBUG2("ifp_firmware_update(): update done. reinitialising device.\n");
	ifp_release(*dh);
	sleep(3);
	return (ifp_init(dh, IRIVER_REINIT));
	DEBUG("ifp_firmware_update(): end\n");
}

/**
 * ifp_download_file:
 * @dh: a usb_dev_handle
 * @iname: the file name on the iRiver do be downloaded 
 * @name: the name of the local file to download to
 * @progress: callback function to signal progress. Set to NULL if no progress
 * indication is needed
 * 
 * Downloads a file (@iname) from the iRiver and stores it in @name.
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_LOCAL_FILE_OPEN_ERROR if 
 * errors happened opening or reading the local file, 
 * %IRIVER_CANNOT_SEND_FILE_NAME if errors communicating with the iRiver
 * happened, %IRIVER_FILE_NOT_FOUND if @name could not be found.
 **/
int ifp_download_file(usb_dev_handle *dh, const char *iname, const char *name, ifp_progress progress) 
{
	int i,size;
	unsigned char ctl[4];
	unsigned char ctl8[8];
	unsigned char buf[0x4000];
	FILE *f;

	DEBUG("ifp_download_file(): begin -- iname=%s name=%s\n", iname, name);
	if (ifp_file_exists(dh,iname) != IRIVER_FILE_EXISTS) 
		return IRIVER_FILE_NOT_FOUND;
	if ((f = fopen(name,"wb")) == NULL) 
		return IRIVER_LOCAL_FILE_OPEN_ERROR;
	if (ifp_send_filename(dh,iname) < 0 ) 
		return IRIVER_CANNOT_SEND_FILENAME;
	usb_req(IFP_FILE_INFO,1,0,ctl);
	DEBUG2("ifp_download_file(): STAGE 1\n");
	usb_control_msg(dh,IFP_REQ_TYPE,IFP_GET_FILE_INFO_SIZE,0,0,ctl8,sizeof(ctl8),TOUT);
	do {
		usb_control_msg(dh,IFP_REQ_TYPE,IFP_FILE_DOWNLOAD,sizeof(buf),0,ctl,sizeof(ctl),TOUT);
		size = ctl[0] + (ctl[1] << 8);
		DEBUG2("ifp_download_file(): STAGE 2\n");
		if (size > 0) {
			if (progress != NULL)
				progress(iname);
			i = usb_bulk_read(dh,IFP_BULK_FROM,buf,size,TOUT);
			fwrite(buf,1,i,f);
		}
	} while (size > 0);
	DEBUG2("ifp_download_file(): STAGE 3\n");
	fclose(f);
	usleep(10000);
	usb_req(IFP_FILE_LEAVE,0,0,ctl);
	DEBUG("ifp_download_file(): end\n");
	return IRIVER_OK;
}


/**
 * ifp_download_dir:
 * @dh: a usb_dev_handle
 * @id: the directory name on the iRiver do be downloaded 
 * @ld: the local path to download to
 * @progress: callback function to signal progress. Set to NULL if no progress
 * indication is needed
 * 
 * Downloads recursively a directory (@id) from the iRiver and stores it in @ld.
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_LOCAL_FILE_OPEN_ERROR if 
 * errors happened opening or reading the local file, 
 * %IRIVER_CANNOT_SEND_FILE_NAME if errors communicating with the iRiver
 * happened, %IRIVER_FILE_NOT_FOUND if @name could not be found.
 **/
int ifp_download_dir(usb_dev_handle *dh, const char *id, const char *ld, ifp_progress progress)
{
	int mode = 0711;
        char *dirname, *iname, *name, *l;
        ifp_file_info *list = NULL, *f = NULL;
	int r;	
	unsigned char ctl[4];
	
	DEBUG("ifp_download_dir(): begin -- id=%s ld=%s\n",id, ld);	
	dirname = ifp_compose_name(ld, basename(id));
	DEBUG2("ifp_download_dir(): mkdir() -- dirname=%s\n",dirname);
	mkdir(dirname, mode);

	DEBUG2("ifp_download_dir(): ifp_ls() -- id=%s\n",id);
	if ((r = ifp_ls(dh, id, &list)) != IRIVER_OK) return r;
	f = list;
	while (list != NULL) 
	{ 
		if (list->is_file)
		{
			iname = ifp_compose_name(id,list->name);
			name = ifp_compose_name(dirname,list->name);
			DEBUG2("ifp_download_dir(): ifp_download_file() -- iname=%s name=%s\n",iname, name);
			r = ifp_download_file(dh,iname, name, progress);
			if ((r == IRIVER_GENERIC_USB_ERROR) || (r == IRIVER_FILE_NOT_FOUND))
			{
				DEBUG2("\033[1mifp_download_dir(): ifp_download_file() returns r=%d\033[0m\n",r);	
				DEBUG2("ifp_download_dir(): bogus call to ifp_get_info()\n");				
				ifp_get_info(dh, iname, &l);
				DEBUG2("ifp_download_dir(): retry ifp_download_file -- iname=%s name=%s\n",iname,name);
				r = ifp_download_file(dh,iname, name, progress);
				DEBUG2("ifp_download_dir(): ifp_download_file() returns r=%d\n",r);	
			}
			if (r != IRIVER_OK) return r;
			free(iname);
			free(name);
		}	
		else
		{
			iname = ifp_compose_name(id,list->name);
			DEBUG2("ifp_download_dir(): ifp_download_dir()\n");
			if ((r = ifp_download_dir(dh, iname, dirname, progress))!=IRIVER_OK) 
				return r;
			free(iname);
		}
		list = list->next;
	}
	ifp_free_file_list(&f);
	free(dirname);
	DEBUG("ifp_download_dir(): end\n");
	return IRIVER_OK;
}


/**
 * ifp_upload_dir:
 * @dh: a usb_dev_handle
 * @id: the directory name on the iRiver to upload to
 * @ld: the local path to upload 
 * @progress: callback function to signal progress. Set to NULL if no progress
 * indication is needed
 * 
 * Upload recursively a local directory (@ld) from the iRiver and stores it in @id.
 * 
 * Returns: %IRIVER_OK if successful, %IRIVER_LOCAL_FILE_OPEN_ERROR if 
 * errors happened opening or reading the local file, 
 * %IRIVER_CANNOT_SEND_FILE_NAME if errors communicating with the iRiver
 * happened, %IRIVER_FILE_NOT_FOUND if @name could not be found.
 **/
int ifp_upload_dir(usb_dev_handle *dh, const char *id, const char *ld, ifp_progress progress)
{
	char *dirname, *iname;
	int r;
int ftw_process(const char *name, const struct stat *status, int type)
{
	if (strcmp(name, ld) == 0) return 0;
	DEBUG2("ifp_upload_dir(): ftw_process(): begin -- name=%s\n",name);
	iname = ifp_compose_name(dirname, name + strlen(ld) + 1);
	if (type == FTW_F)
	{
		DEBUG2("ifp_upload_dir(): ftw_process(): ifp_upload_file() -- iname=%s name=%s\n", iname, name);
		r = ifp_upload_file(dh, iname, name, progress);
		if ((r == IRIVER_CANNOT_SEND_FILENAME) || (r == IRIVER_GENERIC_USB_ERROR))
			{
				DEBUG2("\033[1mifp_upload_dir(): ftw_process(): ifp_upload_file() returns r=%d\033[0m\n",r);
				sleep(2);
				if (r == IRIVER_GENERIC_USB_ERROR) // We can safely ignore that. File gets copied. 
					r = 0;
				else // This one is a bit more tricky.
				{
					DEBUG2("ifp_upload_dir(): ftw_process(): retry ifp_delete_file() and ifp_upload_file()\n");
					r = ifp_delete_file(dh,iname, progress);
					DEBUG2("ifp_upload_dir(): ftw_process(): ifp_delete_file() returns r=%d\n",r);
					ifp_battery_condition(dh);
					r = ifp_upload_file(dh, iname, name, progress);
					DEBUG2("ifp_upload_dir(): ftw_process(): ifp_upload_file() returns r=%d\n",r);
				}
			}
		if (r != IRIVER_OK) return r;
	}
	if (type == FTW_D)
	{
		DEBUG2("ifp_upload_dir(): ftw_process(): ifp_make_dir() - iname=%s\n",iname);
		if ((r = ifp_make_dir(dh,iname)) != IRIVER_OK) return r;

	}	
	free(iname);
	DEBUG2("ifp_upload_dir(): ftw_process(): end\n");
	return 0;
}
	DEBUG("ifp_upload_dir(): begin -- id=%s ld=%s\n",id, ld);
	if (strcmp(id,"/") == 0) strcpy(id,""); /* FIXME: This is broken */
	dirname = ifp_compose_name(id, basename(ld));
	if ((r = ifp_make_dir(dh,dirname)) != IRIVER_OK) return r;
	ftw(ld, ftw_process, 8);
	DEBUG("ifp_upload_dir(): end -- r=%d\n",r);
	return r;
}

/**
 * ifp_battery_condition:
 * @dh: a usb_dev_handle
 * 
 * Checks the battery status of the iRiver
 * 
 * Returns: positive integer if battery is fine.
 **/
/* not tested, not sure: */
int ifp_battery_condition(usb_dev_handle *dh) 
{
	unsigned char ctl8[8];	
	unsigned char bbuf[8];
	int i;
	
	DEBUG2("ifp_battery_condition(): begin\n");
	usb_req(IFP_02_COMMAND,0,IFP_02_BATTERY,ctl8);
	i = usb_bulk_read(dh,IFP_BULK_FROM,bbuf,sizeof(bbuf),TOUT);
	DEBUG2("ifp_battery_condition(); end\n");
	return bbuf[0];
}

/**
 * ifp_firmware_version:
 * @dh: a usb_dev_handle
 * 
 * Returns a number containing the firmware revision of the iRiver
 * 
 * Returns : firmware revision
 **/
float ifp_firmware_version(usb_dev_handle *dh) 
{
	unsigned char ctl8[8];
	unsigned char bbuf[8];
	int i;
	
	DEBUG2("ifp_firmware_version(): begin\n");
	usb_req(IFP_02_COMMAND,0,IFP_02_FIRMWARE,ctl8);
	usb_bulk_read(dh,IFP_BULK_FROM,bbuf,sizeof(bbuf),TOUT);
	i = (bbuf[1]/16)*1000 + (bbuf[1]%16)*100 + (bbuf[0]/16)*10 + (bbuf[0]%16);
	DEBUG2("ifp_firmware_version(): end\n");
	return ((float)i/100);
}

static struct usb_device *ifp_return_dev(struct usb_bus *bus)
{
	static struct usb_device *tmp_dev, *dev = NULL;

	for (bus = usb_busses; bus; bus = bus->next)
		for (tmp_dev = bus->devices; tmp_dev; tmp_dev = tmp_dev->next)
			if ((tmp_dev->descriptor.idVendor == iRiver_Vendor) && 
				(tmp_dev->descriptor.idProduct == iFP_Product_1XX ||
				tmp_dev->descriptor.idProduct == iFP_Product_3XX))
				dev = tmp_dev;
	return(dev);
}

/**
 * ifp_init:
 * @dh: a usb_dev_handle
 * @returns: the error code of the operation
 * @init_type: %IRIVER_INIT or %IRIVER_REINIT
 *
 * This function has to be called before every other ifp_* call
 * as it takes care of finding and initializing the device.
 *
 * Returns: %IRIVER_OK if successful, %IRIVER_NO_USB_BUS if no 
 * usb busses were found, %IRIVER_NO_DEVICE if no usb devices were found,
 * %IRIVER_NO_IFP if no iRiver devices were found, %IRIVER_BUSY is claiming
 * the usb interface was not successful.
 **/
int ifp_init(usb_dev_handle **dh, int init_type)
{
	static struct usb_bus *bus;
	static struct usb_device *tmp_dev, *dev = NULL;
	static struct usb_dev_handle *d;

	DEBUG("ifp_init(): begin -- init_type=%d\n", init_type);
	usb_init();
	if (usb_find_busses() <= init_type) 
		return IRIVER_NO_USB_BUS;
	if (usb_find_devices() <= init_type) 
		return IRIVER_NO_DEVICE;
	if ((dev = ifp_return_dev(bus)) == NULL) 
		return IRIVER_NO_IFP;
	d = usb_open(dev);
	if (usb_claim_interface(d, dev->config->interface->altsetting->bInterfaceNumber) !=0) 
	{
		usb_close(d);
		return IRIVER_BUSY;
	}
	*dh = d;
	DEBUG("ifp_init(): end\n");
	return IRIVER_OK;
}


/**
 * ifp_release:
 * @dh: a usb_dev_handle
 *
 * Releases all the usb resources previously claimed with ifp_init()
 **/
void ifp_release(usb_dev_handle *dh)
{
/* FIXME -- '0' is wrong here. */
	DEBUG("ifp_release(): begin\n");
	usb_release_interface(dh, 0);
	usb_close(dh);
	DEBUG("ifp_release(): end\n");
}

/**
 * ifp_error_to_string:
 * @error: error code
 * 
 * Returns a descriptive string for the error codes returned by the ifp_
 * functions
 * 
 * Returns: a non free-able string
 **/
const char *ifp_error_to_string(int error)
{
	const char *msg[15] = {
		"Cannot open local file! Check the path and permissions.",
		"A generic USB error happened! Check if the device is correctly attached.",
		"Not enough free space on the device to perform this operation!",
		"No USB bus found! Refer to the 'Troubleshooting' section of the README file.",
		"No USB device found! Refer to the 'Troubleshooting' section of the README file.",
		"No IFP player found! Refer to the 'Troubleshooting' section of the README file.",
		"This is not a valid firmware file! Please download the correct firmware file.",
		"Cannot upload file to the IFP. Check if you have enough free space on the device.",
		"Cannot send filename to the device. Check if the filename isn't malformed.",
		"Cannot find the specified file on the device. Please check the path.",
		"The file already exists on the IFP. Please delete it first.",
		"Cannot find the specified directory on the device. Please check the path.",
		"The directory already exists on the IFP. Please delete it first.",
		"The directory is not empty! Please delete all files and directories inside it first.",
		"Cannot claim the device for ourselves. This either means a permission problem or another process is already locking the device."
	};
	return(msg[-(error)-1]);          
}


/**
 * ifp_free_file_list:
 * @first: list to be freed
 * 
 * This functions walks through a whole %ifp_file_info list and frees every
 * element. *first will be set to NULL
 * 
 * Returns: 
 **/
void ifp_free_file_list(ifp_file_info **first)
{
	ifp_file_info *prev, *current;

	current = *first;
	while (current != NULL) {
		free(current->mimetype);
		free(current->name);
		prev=current;
		current=current->next;
		free(prev);
	}
	*first = NULL;
}
